\ Postbox 1.0 - handling of dBase files for RS-MAIL
\ Source text for DX-Forth
\ Pavel Zampach, 2025
\ *

inlib statmem
inlib fileio
inlib applib
inlib buflib
inlib timedate
inlib strvar
inlib dbfconrs
inlib dbfacc
cr ." Postbox 1.0 loading "

decimal
MAILADD nip constant MAIL-LEN
SUBJECT nip constant SUBJ-LEN
ATTACH  nip constant ATTA-LEN
FILE    nip constant FILE-LEN
NAME-A  nip constant NAMEA-LEN
$0169 c@ 1- constant LAST-ROW
$0080 constant SYSBUF
$0d constant _CR
$0169 c@ constant ABOOK-MAX		\ max. records = # of screen rows
recsize-a 1- constant ABOOK-LEN		\ length of ABOOK item
70 constant line-len			\ length of line in msg
variable ACT-REC			\ active record
variable END-FLAG			\ flag for menu end
variable IN-FLAG			\ flag for Inbox/Outbox (true/false)
variable DEL-FLAG			\ flag for delete record
variable ACT-IN				\ active record in RMAIL
variable #ABOOK				\ # records in address book
integer FID-O				\ fid for output file
8 string: DATESTR			\ today date as DD/MM/YY
10 string: MSG-NAME
mail-len string: MAIL-OLD
subj-len string: SUBJ-OLD
file-len string: FILE-OLD
file-len string: FILESTR
abook-max abook-len * buffer: ABOOK-BUF

\ Bits in RMAIL.DBF FLAG field:
\ b0 - Attach (set by RMAIL.COM)
\ b1 - Forward
\ b2 - Reply

system
: [CTRL]	( -- b )
  char $1f and
  postpone literal
; immediate
application

\ Type string [addr u] with Z-System marks (^A, ^B)
: ZTYPE		( addr u -- )
  bounds do
    i c@ cond
      1 of bold endof
      2 of normal endof
      emit
    cont
  loop
;

\ Append number [n] as two digits and "/" to string [addr u1]
: APP2#		( n addr u1 -- addr u2 )
  rot s>d <# # # #> apps
  [char] / appc
;

\ Set today to DATESTR
: SET-DATE	( -- )
  time&date 2000 - -rot swap		\ sec min hr yr mo day --
  <pad app2# app2# rot appn
  datestr s!
  drop 2drop				\ drop sec min hr
;

\ Read active record
: READ-ACT	( -- )
  act-rec @ read-rec
;

\ Write active record
: WRITE-ACT	( -- )
  act-rec @ write-rec
;

\ Set ACT-REC to the last recod and read it
: INIT-ACT	( -- )
  #recs 1- act-rec !
  read-act
;

\ Set ACT-REC from command line and read record
: SET-ACT	( -- )
  cmdtail ?dup if
    /number 2nip d>s act-rec !
    read-act
  else
    drop init-act
  then
;

: HEADER	( -- )
  page
  5 1 at-xy ." ----- Postbox 1.0 -----"
;

\ Common part of .SCREEN-x
: (.SCREEN)	( -- )
  header
  5 3 at-xy bold in-flag @
  if ." INBOX  " else ." OUTBOX  " then
  normal datestr s@ type
  5 5 at-xy ." Message "
  act-rec @ 1+ . [char] / emit #recs .
  8 7 at-xy ." Subject: " SUBJECT type
  11 11 at-xy ." Date: " DATE type 5 spaces
  alive? if ." Msg file: " FILE type then
;

: .DEL		( -- )
  ." \1b[7mDeleted" normal		\ inverse
;

\ Screen for Inbox
: .SCREEN-IN	( -- )
  (.screen)
  11 9 at-xy ." From: " MAILADD type
  25 3 at-xy
  alive? if
    FLAG get-fieldb
    dup 4 and if ." Replied " then
    dup 2 and if ." Forwarded " then
        1 and if ." Attachment" then
  else
    .del
  then
  5 20 at-xy
  s" e\01X\02it, <, >, \01V\02iew, \01D\02elete, \01P\02rint, \01O\02utbox, \01A\02ddress"
  ztype
  5 21 at-xy
  s" \01N\02ew, r\01E\02ply, \01F\02orward, \01R\02eceive, \01S\02end =>"
  ztype
;

\ Screen for Outbox
: .SCREEN-OUT	( -- )
  (.screen)
  13 9 at-xy ." To: " MAILADD type
  5 13 at-xy ." Attachment: " ATTACH type
  25 3 at-xy
  alive? if
    FLAG get-fieldl if
      ." Not sent"
    else
      ." Sent"
    then
  else
    .del
  then
  5 20 at-xy
  s" e\01X\02it, <, >, \01V\02iew, \01D\02elete, \01P\02rint, \01R\02esend, \01I\02nbox =>"
  ztype
;

\ Go to next record
: NEXT-REC	( -- )
  act-rec @ #recs 1- < if
    1 act-rec +!
    read-act
  then
;

\ Go to previous record
: PREV-REC	( -- )
  act-rec @ if
    -1 act-rec +!
    read-act
  then
;

\ Display text from FILE field
: VIEW-REC	( -- )
  page   0				\ line counter
  FILE get-fieldct r/o open-file fc to fid
  begin
    sysbuf line-len fid read-line fc	\ read line
  while
    sysbuf swap type cr			\ display line
    1+	dup last-row = if		\ inc and test counter
      drop
      ." SPACE cont, other end"
      key bl <> if
        fid close-file fc
      end
      _cr emit clear-line
      0					\ reset counter
    then
  repeat   2drop			\ while + line counter
  fid close-file fc
  0 last-row at-xy ." Press any key" key drop
;

\ Print record
: PRINT-REC	( -- )
  printer
  in-flag @ if ." From: " else ." To: " then
  MAILADD type cr
  ." Subject : " SUBJECT type cr
  ." Date: " DATE type cr cr
  FILE get-fieldct r/o open-file fc to fid
  begin
    sysbuf line-len fid read-line fc
  while
    sysbuf swap type cr
  repeat   drop
  fid close-file fc
  page   console
;

\ Delete record
: DEL-REC	( -- )
  5 22 at-xy ." Delete? " y/n if
    false set-alive
    FILE get-fieldct delete-file drop	\ delete msg file, may be missed
    write-act
    del-flag on
  then
;

\ Mark message for send
: RESEND	( -- )
  5 21 at-xy ." Resend? " y/n if
    true FLAG set-fieldl
    write-act
  then
;

\ Switch to Inbox
: INBOX		( -- )
  recsize-r to recsize
  s" RMAIL" open-dbf
  in-flag on
;

\ Switch to Outbox
: OUTBOX	( -- )
  recsize-s to recsize
  s" SMAIL" open-dbf
  in-flag off
;

\ Switch to ABOOK
: ABOOK		( -- )
  recsize-a to recsize
  s" ADDRBOOK" open-dbf
;

\ Display ABOOK items
: .ABOOK	( -- )
  page   abook-buf
  #abook @ 0 do
    3 spaces
    dup abook-len type cr
    abook-len +
  loop   drop
;

\ Return address string [addr u] from ABOOK
: ABOOK-PICK	( -- addr u )
  .abook
  0					\ init pointer
  begin
    0 over at-xy ." =>"
    key >r
    _cr emit space space
    r@ [ctrl] E = over and if 1- then
    r@ [ctrl] X = over #abook @ 1- < and if 1+ then
    r> _cr =
  until
  abook-len * abook-buf + namea-len +   mail-len
;

\ Init ABOOK, no dBase file is opened
: ABOOK-INIT	( -- )
  abook
  #recs #abook !
  abook-buf
  #recs 0 do
    i read-rec
    recbuff 1+ over abook-len cmove
    abook-len +
  loop   drop
  close-dbf
;

\ Build file for reply or forward
: QUOT-FILE	( -- )
  file-old s@ r/o open-file fc to fid
  filestr s@ w/o create-file fc to fid-o
  s" \0d\0a\0d\0a" fid-o write-file fc	\ two blank lines
  begin
    pad line-len fid read-line fc
  while
    s" > " fid-o write-file fc
    pad swap fid-o write-line fc
  repeat   drop
  fid-o close-file fc
  fid close-file fc
;

\ Header for MSG-INPUT
: MSG-I-HEAD	( -- )
  header
  5 5 at-xy msg-name s@ type space ." message"
  13 7 at-xy ." To :"
;

\ [addr u] is string with subject prefix
: SET-SUBJ	( addr u -- )
  >pad subj-old s@ apps SUBJECT set-fieldcl
;

\ [u] is mask for bit settings
: SET-FLAG	( u -- )
  ?dup if
    FLAG get-fielda dup c@ rot or swap c!
    write-act
  then
;

\ Input fields. [n] is bit mask for fields enter
: MSG-INPUT	( n -- )
  dup 1 and if
    msg-i-head
    pad dup mail-len accept
    dup 0= if
      2drop   abook-pick
    then
    MAILADD set-fieldcl
  then
  msg-i-head MAILADD type
  8 9 at-xy ." Subject :"
  dup 2 and if
    pad dup subj-len accept
    SUBJECT set-fieldcl
  else
    SUBJECT type
  then
  5 11 at-xy ." Attachment: "
  pad dup atta-len accept
  ATTACH set-fieldcl
;

\ New message
: MSG-NEW	( -- )
  s" New" msg-name s! 3 msg-input
;

\ Replied message
: MSG-REPLY	( -- )
  quot-file
  mail-old s@ MAILADD set-fieldcl
  s" Re: " set-subj
  s" Replied" msg-name s! 0 msg-input
;

\ Forwarded message
: MSG-FORW	( -- )
  quot-file
  s" Fw: " set-subj
  s" Forwarded" msg-name s! 1 msg-input
;

: STORE-OLD	( -- )
  MAILADD get-fieldct mail-old s!	\ store orig. mail address
  SUBJECT get-fieldct subj-old s!	\ store orig. subject
  FILE get-fieldct file-old s!		\ store orig. filename
  act-rec @ act-in !			\ store ACT-REC from RMAIL.DBF
;

\ For ZDE is needed to terminate program by jump to address 0
code EXIT-PRG	( -- )
	0	jmp
end-code

\ Build command line [addr u] and exit program
: CMDLINE	( addr u -- )
  page close-dbf
  zenv 24 + @				\ addr of command line buffer
  dup 2+ 2+ dup rot !			\ write addr of the first char
  0					\ command line as empty string
  +string				\ add command
  + off exit-prg			\ add trailing null and exit
;

\ Set FILESTR as "MO####.MSG", [addr u] is content of FILE field
: SET-FILESTR	( addr u -- )
  2 /string /number 2nip 1 m+
  <# s" .MSG" shold # # # # s" MO" shold #>
  filestr s!
;

\ [xt] is routine for fields settings, [u] is mask for FLAG
: MAKE-MSG	( xt u -- )
  set-flag store-old
  close-dbf
  outbox init-act
  FILE get-fieldct set-filestr
  execute				\ set MAILADD, SUBJECT and ATTACH
  true FLAG set-fieldl
  datestr s@ DATE set-fieldcl
  filestr s@ FILE set-fieldcl
  append-rec
  s" D1:ZDE " >pad filestr s@ apps
  s" /A;POSTBOX " apps act-in @ appn
  cmdline
;

\ Add MAILADR from actual msg to ABOOK
: ABOOK-ADD	( -- )
  MAILADD get-fieldct mail-old s!	\ store mail address
  close-dbf
  abook
  #recs abook-max < if
    mail-old s@ MAIL-A set-fieldcl
    5 23 at-xy ." Name: "
    pad dup namea-len accept
    NAME-A set-fieldcl
    append-rec
  then
  close-dbf
  abook-init
  inbox read-act
;

: MENU		( -- )
  begin
    in-flag @ if .screen-in else .screen-out then
    key upcase dup
    case
      [ctrl] D of next-rec endof
      [char] . of next-rec endof
      [ctrl] S of prev-rec endof
      [char] , of prev-rec endof
      [char] X of end-flag on endof
      [char] V of alive? if view-rec then endof
      [char] D of del-rec endof
      [char] P of alive? if print-rec then endof
      [char] A of abook-add endof
    endcase
    in-flag @ if
      case
	[char] O of close-dbf outbox init-act endof
	[char] N of ['] msg-new 0 make-msg endof
	[char] E of alive? if ['] msg-reply 4 make-msg then endof
	[char] F of alive? if ['] msg-forw 2 make-msg then endof
	[char] R of s" RMAIL;POSTBOX" cmdline endof
	[char] S of s" SMAIL;POSTBOX" cmdline endof
      endcase
    else
      case
	[char] I of close-dbf inbox init-act endof
	[char] R of alive? if resend then endof
      endcase
    then
    end-flag @
  until
;

: INIT		( -- )
  set-date
  del-flag off
  end-flag off
  abook-init
  inbox set-act
;

: MAIN		( -- )
  init
  menu
  del-flag @ if
    s" DBASE PACK" cmdline
  then
  close-dbf
;

[DEFINED] allocate [IF]
turnkey main e0:postbox
[THEN]
