\ Send mail to SMTP server according data in SMAIL.DBF
\ Part of RS-mail system
\ Ver. 2.1 (partial MIME support)
\ Pavel Zampach, 2022
\ *

inlib fileio
inlib statmem				\ static memory management
inlib buflib				\ buffer:
inlib strcon				\ string constant
inlib strvar				\ string variable
inlib applib				\ string appending
inlib timedate				\ TIME&DATE
inlib base64				\ Base64 en/decoding
inlib dbfcons				\ field constants for SMAIL.DBF
inlib dbfacc				\ DBF file access
inlib modem				\ modem
inlib ndr				\ NDR filename conversion
inlib dstlib				\ DST correction
inlib ccq				\ C,"
cr ." SMAIL 2.1 loading"

decimal

char > constant ">"
char : constant ":"
char " constant QUOT
s" Content-Type: text/plain" constr PLAIN
s" --frontier" constr BOUND		\ must match with next line!
s" Content-Type: multipart/mixed; boundary=\22frontier\22" constr MULTI
s" Content-Transfer-Encoding: " constr TENCO
54 constant BUF-SIZE			\ for base64 standard line length
128 buffer: BUF
max-path string: ATT-NAME		\ attachment filename
variable CONNECTION			\ flag for connection to SMTP server

\ Mail account specific data
s" mail.domain.com:25" constr MCONN	\ URL of smtp server
s" EHLO domain.com" constr MDOMA	\ domain
s" name@domain.com" constr MADDR	\ sender's mail address
aka maddr MUSER				\ login name (mail address in this case)
s" password" constr MPASS		\ password
s" +0100" constr TZ0			\ basic timezone (CET)
s" +0200" constr TZ1			\ timezone with DST correction


\ Return the first substring [addr3 u3] from string [addr1 u1]
\ separated by char [c], string [addr2 u2] is the remainder. When
\ [addr3 u3] is the last part of original string, u2 = -1
: SLICE		( addr1 u1 c  -- addr2 u2 addr3 u3 )
  >r 2dup r> scan dup >r
  1 /string 2swap r> -
;

\ Duplicate a triple number
: 3DUP		( x1 x2 x3 -- x1 x2 x3 x1 x2 x3 )
  dup 2over rot
;

\ End of session with [addr u] msg
: END-SES	( addr u -- )
  type cr
  flushport
  s" QUIT" send-line flushport		\ QUIT and read the rest of data
  mreset
  quit
;

\ Read chars to PAD, end by timeout
: READ-TIME	( -- )
  pad
  begin
    skeysec if 2drop end		\ timeout -> end
    over c! 1+				\ store char to memory
    pad 80 + min			\ trim addr
  again
;

\ Extended timeout check
:NONAME		( ior --- )
  if s" Time out" end-ses then
; is tc

\ Send string [addr u] to modem, read and check response
: SEND+CHECK	( addr u -- )
  send-line
  read-time
  pad c@ [char] 4 > if
    ." Server error " pad 3 type cr
    end-ses
  then
;

\ Connect and login to SMTP server
: CONNECT	( -- )
  connection @ if end

  flushport
  mreset
  mconn mdial
  500 to wticks
  begin
    pad rcv-line
    pad 1+ c@ [char] 2 =		\ wait for message "2xx ..."
  until
  20 to wticks
  mdoma send+check
  s" AUTH LOGIN" send+check
  <pad muser app-base64e send+check
  <pad mpass app-base64e send+check
  ." Connected" cr
  connection on
;

\ Single dot as mark of msg end
: MSG-END	( -- )
  s" ." send+check
;

\ Attachment filename w/o path [addr u]
: ATT-NAME-	( -- addr u )
  att-name s@ -path
;

\ [flag] is TRUE if attachment is present
: ATTACH?	( -- flag )
  ATTACH get-fieldct nip
;

create TAB-MO c," JanFebMarAprMayJunJulAugSepOctNovDec"

\ No of month (1..12) -> string [addr 3]
: MO>STRING	( n -- addr 3 )
  1- 3 * tab-mo + 3
;

\ Append [n] as two digits string to string [addr u1]
: APP2DIG	( addr u1 n -- addr u2 )
  s>d <# # # #> apps
;

\ Date for msg header (in BUF)
\ format: 5 Jan 2022 19:23:00 +0100
: GET-DATE	( -- addr u )
  time&date 3dup dst >r			\ save DST correction
  swap rot				\ s min h d mo y -- s min hr y mo d
  buf 0					\ empty string in BUF
  rot appn appbl			\ day
  rot mo>string apps appbl		\ month
  rot appn appbl			\ year
  rot app2dig ":" appc			\ hour
  rot app2dig ":" appc			\ min
  rot app2dig appbl			\ sec
  r> if tz1 else tz0 then apps		\ timezone
;

\ Send msg header
: SEND-HEADER	( -- )
  \ commands for SMTP server
  s" MAIL FROM: " >pad maddr apps send+check
  s" RCPT TO: " >pad MAILADD get-fieldct apps send+check
  s" DATA" send+check
  \ message data follows:
  s" From: <" >pad maddr apps ">" appc send-line
  s" To: <" >pad MAILADD get-fieldct apps ">" appc send-line
  s" Subject: " >pad SUBJECT get-fieldct apps send-line
  s" Date: " >pad get-date apps send-line
  s" MIME-version: 1.0" send-line
  attach? if multi else plain then send-line
  s" X-Mailer: CP/M RS-mail" send-line
  send-empty				\  mark of header end
;

\ Internal component generating exception
: (SEND-TXT)	( -- )
  FILE get-fieldct r/o open-file throw to fid
  begin
    pad 80 fid read-line throw
  while
    pad swap send-line
  repeat   drop
  fid close-file throw
;

\ Internal component generating exception
: (SEND-B64)   ( -- )
  att-name s@ r/o open-file throw to fid
  begin
    buf buf-size fid read-file throw dup
  while
    >r <pad buf r> app-base64e send-line
  repeat   drop
  fid close-file throw
;

\ Run [xt] and handle exceptions
: LAUNCH	( xt -- )
  catch if
    ." File error" cr
    fid close-file drop
  then
;

\ Send msg as plain text
: SEND-TXT	( -- )
  ['] (send-txt) launch
;

\ Send attachment Base64 encoded
: SEND-B64	( -- )
  ['] (send-b64) launch
;

: SEND-FILE	( -- )
  bound send-line
  s" Content-Type: application/octet-stream; name=\22" >pad
  att-name- apps quot appc send-line
  tenco >pad s" base64" apps send-line
  s" Content-Disposition: attachment; filename=\22" >pad
  att-name- apps quot appc send-line
  send-empty				\ mark of header end
  send-b64
;

\ Send msg body with attachment(s)
: SEND-ATT	( -- )
  \ send msg text
  bound send-line
  plain send-line
  tenco >pad s" 7bit" apps send-line
  send-empty				\ mark of header end
  send-txt
  \ send attachment(s)
  ATTACH get-fieldct 2dup upper
  begin
    [char] ; slice -blanks -ndr att-name s!
    send-file
    dup 0<				\ if was the last filename, end
  until   2drop
  bound >pad s" --" apps send-line
;

\ Send whole message
: SEND-MSG	( -- )
  connect
  ." Sending mail to " MAILADD get-fieldct type
  send-header
  attach? if send-att else send-txt then
  msg-end
  ." \20done" cr
;

\ Send all messages from SMAIL.DBF where field FLAG is TRUE
: MAIN		( -- )
  connection off
  s" SMAIL" open-dbf
  #recs 0 ?do
    i read-rec
    FLAG get-fieldl alive? and if
      send-msg
      false FLAG set-fieldl
      i write-rec
    then
  loop
  close-dbf
  connection @ if s" Quit" end-ses then
;

turnkey main e0:smail
